# German: Soll -> Abgang vom Konto
#        Haben -> Zugang aufs Konto
# English: Debit -> from account
#         Credit -> to account

from sqlalchemy import func
from datetime import datetime, timezone
from werkzeug.exceptions import BadRequest, NotFound

from flaschengeist.database import db
from flaschengeist.models.user import User

from .models import Transaction
from . import permissions
from ... import logger

__attribute_limit = "balance_limit"


def set_limit(user: User, limit: float, override=True):
    if override or not user.has_attribute(__attribute_limit):
        user.set_attribute(__attribute_limit, limit)
        db.session.commit()


def get_limit(user: User) -> float:
    return user.get_attribute(__attribute_limit, default=None)


def get_balance(user, start: datetime = None, end: datetime = None):
    if not start:
        start = datetime.fromtimestamp(0, tz=timezone.utc)
    if not end:
        end = datetime.now(tz=timezone.utc)

    credit = (
        db.session.query(func.sum(Transaction.amount))
        .filter(Transaction.receiver_ == user)
        .filter(Transaction._reversal_id.is_(None))  # ignore reverted transactions
        .filter(start <= Transaction.time)
        .filter(Transaction.time <= end)
        .scalar()
    ) or 0

    debit = (
        db.session.query(func.sum(Transaction.amount))
        .filter(Transaction.sender_ == user)
        .filter(Transaction._reversal_id.is_(None))  # ignore reverted transactions
        .filter(start <= Transaction.time)
        .filter(Transaction.time <= end)
        .scalar()
    ) or 0

    return credit, debit, credit - debit


def send(sender: User, receiver, amount: float, author: User):
    """Send credit from one user to an other

    Args:
        sender: User who sends the amount
        receiver: User who receives the amount
        amount: Amount to send
        author: User authoring this transaction
    Returns:
        Transaction that was created
    Raises:
        BadRequest if amount <= 0
    """
    if amount <= 0:
        raise BadRequest

    if sender and sender.has_attribute(__attribute_limit):
        if (get_balance(sender)[2] - amount) < sender.get_attribute(__attribute_limit) and not author.has_permission(
            permissions.EXCEED_LIMIT
        ):
            raise BadRequest("Limit exceeded")

    transaction = Transaction(sender_=sender, receiver_=receiver, amount=amount, author_=author)
    db.session.add(transaction)
    db.session.commit()
    return transaction


def change_balance(user, amount: float, author):
    """Change balance of user

    Args:
        user: User to change balance
        amount: Amount to change balance
        author: User authoring this transaction
    """
    sender = user if amount < 0 else None
    receiver = user if amount > 0 else None
    return send(sender, receiver, abs(amount), author)


def get_transaction(transaction_id) -> Transaction:
    transaction = Transaction.query.get(transaction_id)
    if not transaction:
        raise NotFound
    return transaction


def reverse_transaction(transaction: Transaction, author: User):
    """Reverse a transaction

    Args:
        transaction: Transaction to reverse
        author: User that wants the transaction to be reverted
    """
    reversal = send(transaction.receiver_, transaction.sender_, transaction.amount, author)
    reversal.reversal = transaction
    transaction.reversal = reversal
    db.session.commit()
    return reversal
